--local DEBUG = true
EventHorizon = {}
EventHorizonDB = {
	point = {'CENTER', 'UIParent', 'CENTER'},
	isActive = true,
	version = 1,
}
local EventHorizon = EventHorizon
EventHorizon.config = {}
EventHorizon.barconfigs = {}
EventHorizon.barframes = {}
EventHorizon.glyphs = {}
EventHorizon.otherIDs = {}
EventHorizon.db = EventHorizonDB

local UnitDebuff = UnitDebuff

local eventhandler
local EventHorizonFrame = CreateFrame('Frame','EventHorizonFrame',UIParent)
local mainframe = CreateFrame('Frame',nil,EventHorizonFrame)
mainframe:SetAllPoints()
EventHorizon.mainframe = mainframe

local activeFrames = {}

local playerguid
local onepixelwide = 1

local function printhelp(...) if select('#',...)>0 then return tostring((select(1,...))), printhelp(select(2,...)) end end
local function debug(...)
	print(...)
end
local function print(...)
	ChatFrame1:AddMessage('EventHorizon: '..strjoin(',',printhelp(...)))
end

local reloadEvents = {
	['COMBAT_LOG_EVENT_UNFILTERED'] = true,
	['PLAYER_TALENT_UPDATE'] = true,
	['UPDATE_SHAPESHIFT_FORM'] = true,
	['UPDATE_SHAPESHIFT_FORMS'] = true,
	['GLYPH_ADDED'] = true,
	['GLYPH_ENABLED'] = true,
	['GLYPH_REMOVED'] = true,
	['GLYPH_UPDATED'] = true,
	['GLYPH_DISABLED'] = true,
	['SPELL_UPDATE_COOLDOWN'] = true,
	['ZONE_CHANGED_NEW_AREA'] = true,
	['ZONE_CHANGED'] = true,
	['ZONE_CHANGED_INDOORS'] = true,
	['PLAYER_ALIVE'] = true,
	['PLAYER_REGEN_DISABLED'] = true,
	['PLAYER_REGEN_ENABLED'] = true,
}

-- Dispatch event to method of the event's name.
local function EventHandler(self, event, ...)
	local f = self[event]
	if f then 
		f(self,...) 
	end 
end

local function Clone(t)
	local new = {}
	local i, v = next(t, nil)	-- i is an index of t, v = t[i]
	while i do
		new[i] = v
		i, v = next(t, i)
	end
	return new
end

-- Helper function for unique debuffs that accepts a list of aura names.
-- Intended usage is for unique debuffs like Mangle - Cat/Mangle - Bear/Trauma, which overwrite each other.
local function UnitDebuffUnique(unitid, auras, unique)
	if unique == true then
		for i,auraname in ipairs(auras) do
			local name, rank, icon, count, debuffType, duration, expirationTime, source = UnitDebuff(unitid, auraname)
			if name then
				return name, rank, icon, count, debuffType, duration, expirationTime, source
			end
		end
	else
		for i,auraname in ipairs(auras) do
			local name, rank, icon, count, debuffType, duration, expirationTime, source = UnitDebuff(unitid, auraname)
			if name and (source == 'player') then
				return name, rank, icon, count, debuffType, duration, expirationTime, source
			end
		end
	end
end

local function UnitBuffUnique(unitid, auras, unique)
	if unique == true then
		for i,auraname in ipairs(auras) do
			local name, rank, icon, count, debuffType, duration, expirationTime, source = UnitBuff(unitid, auraname)
			if name then
				return name, rank, icon, count, debuffType, duration, expirationTime, source
			end
		end
	else
		for i,auraname in ipairs(auras) do
			local name, rank, icon, count, debuffType, duration, expirationTime, source = UnitBuff(unitid, auraname)
			if name and (source == 'player') then
				return name, rank, icon, count, debuffType, duration, expirationTime, source
			end
		end
	end
end

--[[
Class SpellFrame
All spell bar frames inherit from this class.
--]]
local SpellFrame = {}
function SpellFrame:NotInteresting(unitid, spellname) 
	return unitid ~= 'player' or spellname ~= self.spellname
end

--[[
Indicators represent a point or range of time. There are different types. The type determines the color and position.
--]]
local colors = {
	sent = {1,1,1,1},
	tick = {1,1,1,1},
	casting = {0,1,0,0.3},
	cooldown = {1,1,1,0.3},
	debuffmine = {1,1,1,0.3},
	debuff = {0.6,0.6,0.6,0.3},
	playerbuff = {0,1,0,0.3},
	default = {1,1,1,1},
}
local layouts = {
--[[
	cooldown = {
		top = {'TOP', 0, -6},
		bottom = {'BOTTOM', 0, 3},
	},
]]--
	tick = {
		top = {'TOP',0,0},
		bottom = {'TOP', 0, -5},
	},
	frameline = {
		top = {'TOP',0,0},
		bottom = {'BOTTOM',0,0},
	},
	smalldebuff = {
		top = {'TOP', 0, -3},
		bottom = {'TOP', 0, -6},
	},
	cantcast = {
		top = {'TOP', 0, -6},
	},
	default = {
		top = {'TOP', 0, -5},
		bottom = {'BOTTOM', 0, 0},
	},
}

local past, future
local barwidth, barheight, barspacing, scale
local bartexture, barbgcolor
local texturedbars

local classburn
local classalpha

local typeparent = {}
local drawingorder = {'default', 'debuff', 'debuffmine', 'playerbuff', 'cooldown', 'casting', 'sent', 'tick','now'}

EventHorizon.colors = colors
EventHorizon.layouts = layouts

function SpellFrame:AddIndicator(typeid, layoutid, time, usetexture)
	local indicator
	local parent = typeparent[typeid]
	if not parent then
		parent = CreateFrame("Frame", nil, mainframe)
		local framelevel = parent:GetFrameLevel()
		for i,t in ipairs(drawingorder) do
			if t==typeid then
				framelevel = framelevel+i
				break
			end
		end
		parent:SetFrameLevel(framelevel)
		parent.unused = {}
		typeparent[typeid] = parent
		--if DEBUG and typeid=='tick'  then parent.numchildren=0 end--]]
	end
	if #parent.unused>0 then
		indicator = tremove(parent.unused)
		indicator:ClearAllPoints()
		indicator.time = nil
		indicator.start = nil
		indicator.stop = nil
		indicator.happened = nil
		--if DEBUG and typeid=='tick'  then debug('reusing indicator',indicator.frameindex) end--]]
	else
		indicator = parent:CreateTexture(nil, 'ARTWORK')
		indicator.parent = parent
		--if DEBUG and typeid=='tick' then parent.numchildren=parent.numchildren+1 indicator.frameindex=parent.numchildren debug('adding indicator',indicator.frameindex) end--]]
	end
	-- Layout
	local layout = layouts[layoutid] or layouts.default
	indicator:SetPoint('TOP',layoutid == 'frameline' and EventHorizon.mainframe or self,unpack(layout.top))
	indicator:SetPoint('BOTTOM',layoutid == 'frameline' and EventHorizon.mainframe or self,unpack(layout.bottom))

	local color = colors[typeid] or colors.default
	if usetexture then
		indicator:SetTexture(bartexture)
		indicator:SetTexCoord(unpack(layout.texcoords))
	else
		indicator:SetTexture(1,1,1,1)
	end
	indicator:SetVertexColor(unpack(color))

	indicator:Hide()
	indicator:SetWidth(onepixelwide)
	indicator.time = time
	indicator.typeid = typeid
	if indicator then
		tinsert(self.indicators, indicator)
	end
	return indicator
end

function SpellFrame:AddSegment(typeid, layoutid, start, stop)
	if stop<start then return end
	local indicator = self:AddIndicator(typeid, layoutid, start, texturedbars)
	indicator.time = nil
	indicator.start = start
	indicator.stop = stop
	--debug(start,stop)
	return indicator
end

function SpellFrame:Remove(indicator)
	if type(indicator)=='number' then
		local index, indicator = indicator, self.indicators[indicator]
		indicator:Hide()
		--if DEBUG and indicator.typeid=='tick' then debug('deleting',indicator.frameindex) end--]]
		tinsert(indicator.parent.unused, tremove(self.indicators, index))
	else
		for index=1,#self.indicators do
			if self.indicators[index]==indicator then
				indicator:Hide()
				--if DEBUG and indicator.typeid=='tick' then debug('deleting',indicator.frameindex) end--]]
				tinsert(indicator.parent.unused, tremove(self.indicators,index))
				break
			end
		end
	end
end

function SpellFrame:OnUpdate(elapsed)
	local now = GetTime()
	local diff = now+past

	-- spellframe.nexttick is used to schedule the creation of predicted ticks as soon as they scroll past now+future.
	local nexttick = self.nexttick
	if nexttick and nexttick <= now+future then
		if nexttick<=self.lasttick then
			self:AddIndicator('tick', 'tick', nexttick)
			self.latesttick = nexttick
			self.nexttick = nexttick + (self.dotMod or self.dot)
		else
			self.nexttick = nil
		end
	end
	for k=#self.indicators,1,-1 do
		local indicator = self.indicators[k]
		local time = indicator.time
		if time then
			-- Example: 
			-- [-------|------->--------]
			-- past    now     time     future
			-- now=795, time=800, past=-3, then time is time-now-past after past.
			local p = (time-diff)*scale
			local remove = p<0 or (time<=now and indicator.typeid=='tick' and not indicator.happened)
			if remove then
				indicator:Hide()
				--if DEBUG and indicator.typeid=='tick' then debug('deleting',indicator.frameindex) end--]]
				tinsert(indicator.parent.unused, tremove(self.indicators,k))
			elseif p<=1 then
				indicator:SetPoint('LEFT', self, 'LEFT', p*barwidth, 0)
				indicator:Show()
			end
		else
			local start, stop = indicator.start, indicator.stop
			local p1 = (start-diff)*scale
			local p2 = (stop-diff)*scale
			if p2<0 then
				indicator:Hide()
				--if DEBUG and indicator.typeid=='tick' then debug('deleting',indicator.frameindex) end--]]
				tinsert(indicator.parent.unused, tremove(self.indicators,k))
			elseif 1<p1 then
				indicator:Hide()
			else
				indicator:Show()
				indicator:SetPoint('LEFT', self, 'LEFT', 0<=p1 and p1*barwidth or 0, 0)
				indicator:SetPoint('RIGHT', self, 'LEFT', p2<=1 and p2*barwidth+1 or barwidth, 0)
			end
		end
	end
end
-- /run e.mf:AddIndicator(GetTime())

function SpellFrame:UNIT_SPELLCAST_SENT(unitid, spellname, spellrank, spelltarget)
	--debug('UNIT_SPELLCAST_SENT',unitid, spellname, spellrank, spelltarget)
	local now = GetTime()
	if self:NotInteresting(unitid, spellname) then return end
	self:AddIndicator('sent', 'default', now)
end

function SpellFrame:UNIT_SPELLCAST_CHANNEL_START(unitid, spellname, spellrank)
	if self:NotInteresting(unitid, spellname) then return end
	local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill = UnitChannelInfo(unitid)
	if not startTime or not endTime then return end
	startTime, endTime = startTime/1000, endTime/1000
	self.casting = self:AddSegment('casting', 'default', startTime, endTime)
	if self.numhits then
		local casttime = endTime - startTime
		local tick = casttime/self.numhits
		for i=1,self.numhits do
			self:AddIndicator('tick', 'tick', startTime + i*tick)
		end
	end
	if (EventHorizon.config.castLine == true) and (endTime - startTime > 1.5) then
		self.castLine = self:AddIndicator('sent', 'frameline', endTime)
	end
end

function SpellFrame:UNIT_SPELLCAST_CHANNEL_UPDATE(unitid, spellname, spellrank)
	--debug('UNIT_SPELLCAST_CHANNEL_UPDATE',unitid, spellname, spellrank)
	if self:NotInteresting(unitid, spellname) then return end
	local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill = UnitChannelInfo(unitid)
	if not startTime or not endTime then return end
	startTime, endTime = startTime/1000, endTime/1000
	if self.casting then
		self.casting.stop = endTime
		if (EventHorizon.config.castLine == true) and self.castLine then
			self.castLine.time = endTime
		end
	end
	self:RemoveTicksAfter(endTime)
end

function SpellFrame:UNIT_SPELLCAST_CHANNEL_STOP(unitid, spellname, spellrank)
	if self:NotInteresting(unitid, spellname) then return end
	local now = GetTime()
	if self.casting then
		self.casting.stop = now
		if EventHorizon.config.castLine == true and self.castLine then
			self.castLine.time = now
		end
		self.casting = nil
	end
	self:RemoveTicksAfter(now)
end

function SpellFrame:UNIT_SPELLCAST_START(unitid, spellname, spellrank, target)
	if self:NotInteresting(unitid, spellname) then return end
	local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill = UnitCastingInfo(unitid)
	if not startTime or not endTime then return end
	startTime, endTime = startTime/1000, endTime/1000
	self.casting = self:AddSegment('casting', 'default', startTime, endTime)
	if (EventHorizon.config.castLine == true) and (endTime - startTime > 1.5) then
		self.castLine = self:AddIndicator('sent', 'frameline', endTime)
	end
end

function SpellFrame:UNIT_SPELLCAST_STOP(unitid, spellname, spellrank)
	local now = GetTime()
	if self:NotInteresting(unitid, spellname) then return end
	if self.casting then
		self.casting.stop = now
		if (EventHorizon.config.castLine == true) and self.castLine then
			self.castLine.time = now
		end
		self.casting = nil
	end
end

function SpellFrame:UNIT_SPELLCAST_DELAYED(unitid, spellname, spellrank)
	--debug('UNIT_SPELLCAST_CHANNEL_UPDATE',unitid, spellname, spellrank)
	if self:NotInteresting(unitid, spellname) then return end
	local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill = UnitCastingInfo(unitid)
	if not startTime or not endTime then return end
	startTime, endTime = startTime/1000, endTime/1000
	if self.casting.stop then
		self.casting.stop = endTime
		if EventHorizon.config.castLine == true and self.castLine then
			self.castLine.time = endTime
		end
	end
end

function SpellFrame:UNIT_SPELLCAST_SUCCEEDED(unitid, spellname, spellrank)
	if self:NotInteresting(unitid, spellname) then return end
	self.succeeded = GetTime()
end


function SpellFrame:UNIT_AURA(unitid)
	if unitid~=self.auraunit then return end
	local name, rank, icon, count, debuffType, duration, expirationTime, source --= self.AuraFunction(self.auraunit, self.auraname)
	if type(self.auraname) == 'table' then
		if self.unique then
			name, rank, icon, count, debuffType, duration, expirationTime, source = self.AuraFunction(self.auraunit, self.auraname, true)
		else
			name, rank, icon, count, debuffType, duration, expirationTime, source = self.AuraFunction(self.auraunit, self.auraname)
		end
	else
		name, rank, icon, count, debuffType, duration, expirationTime, source = self.AuraFunction(self.auraunit, self.auraname)
	end
		
	local afflicted = name and (self.unique or source=='player') and (not self.minstacks or count>=self.minstacks)
	local addnew
	local now = GetTime()
	local start
	
	local targ = UnitName(self.auraunit)
	
	if self.textureID and self.textureID ~= icon then
		return
	end
		
	
	if self.aurasegment and expirationTime == 0 and duration == 0 then	-- Timeless aura, bar exists (Overkill)
		for i = #self.indicators,1,-1 do
			self:Remove(i)
		end
		self.aurasegment = nil
		self.nexttick = nil
		self.stacks:SetText()
		return
	end
	
	if expirationTime == 0 then
		return
	end
	
	if afflicted then
		start = expirationTime-duration
		if self.aurasegment and (self.aurasegment.lastunit == targ) then
			-- The aura is currently displayed
			if expirationTime~=self.aurasegment.stop then
				if self.alwaysrefresh and not self.cast then -- alwaysrefresh = buff. Cast + buff - HoT = BAD. Buffs with cast time and no HoT component are treated much differently.
					if self.dot then -- ...check to see if it's a HoT. If so, it's treated as a DoT.
						self.aurasegment.stop = start-0.2
						if self.cast and self.useSmalldebuff then
							self.cantcast.stop = start-0.2
						end
						self:RemoveTicksAfter(start)
						addnew = true
						if self.internalcooldown then		-- If we have an internal cooldown set for an aura or trinket proc, display it
							local ready = now + self.internalcooldown
							self:AddSegment('cooldown', 'cooldown', expirationTime, ready)
						end
					else
						-- If it's a buff with no cast time or HoT component, no special handling needed, move along.
						self.aurasegment.stop = expirationTime
						if self.internalcooldown then		-- If we have an internal cooldown set for an aura or trinket proc, display it
							local ready = now + self.internalcooldown
							self:AddSegment('cooldown', 'cooldown', expirationTime, ready)
						end
					end
				else
					-- The aura was replaced.
					self.aurasegment.stop = start-0.2
					if self.cast and self.useSmalldebuff then
						self.cantcast.stop = start-0.2
					end
					self:RemoveTicksAfter(start)
					addnew = true
					if self.internalcooldown then		-- If we have an internal cooldown set for an aura or trinket proc, display it
						local ready = now + self.internalcooldown
						self:AddSegment('cooldown', 'cooldown', expirationTime, ready)
					end
				end
			end
		else
			addnew = true
			if self.internalcooldown then		-- If we have an internal cooldown set for an aura or trinket proc, display it
				local ready = now + self.internalcooldown
				self:AddSegment('cooldown', 'cooldown', expirationTime, ready)
			end
		end
		if count>0 then
			self.stacks:SetFormattedText('%d',count)
		end
	else
		if self.aurasegment then
			if math.abs(self.aurasegment.stop - now)>0.3 then
				self.aurasegment.stop = now
				if self.cast and self.useSmalldebuff then
					self.cantcast.stop = now-0.2
				end
			end
			self:RemoveTicksAfter(now)
			self.aurasegment = nil
			self.stacks:SetText()
		end
	end
	self:UpdateDoT(addnew, source, now, start, expirationTime, name)
end

local typeids = {
	['tick'] = true,
	['cantcast'] = true,
	['debuff'] = true,
	['playerbuff'] = true,
	['debuffmine'] = true,
}
function SpellFrame:PLAYER_TARGET_CHANGED()
	if self.auraunit ~= 'target' then return end
	if self.aurasegment then
		for i = #self.indicators,1,-1 do
			local ind = self.indicators[i]
			if typeids[ind.typeid] then
				self:Remove(i)
			end
		end
		self.aurasegment = nil
		self.nexttick = nil
		self.stacks:SetText()
	end

	if UnitExists('target') then
		self:UNIT_AURA('target')
	end
end

--[[
Refreshable debuffs are really ugly, because UnitDebuff won't tell us when the debuff was applied.
It gets complicated because a debuff might be [re]applied/refreshed when we're not looking.
Here's what can happen, and which point of time we have to assume as the start.
success applied -> applied
success applied refresh -> applied
applied success -> success
applied success refresh -> success
applied -> applied
applied refresh -> applied

So we need to keep track of every debuff currently applied, and the success events. 
When the target or debuff changes, we need to look at the time of the last success to see if we can trust UnitDebuff.
--]]

local tickevents = {
	['SPELL_PERIODIC_DAMAGE'] = true,
	['SPELL_PERIODIC_HEAL'] = true,
	['SPELL_PERIODIC_ENERGIZE'] = true,
	['SPELL_PERIODIC_DRAIN'] = true,
	['SPELL_PERIODIC_LEACH'] = true,	
	['SPELL_DAMAGE'] = true,
	['SPELL_HEAL'] = true,
	--['SPELL_AURA_APPLIED'] = true,
}

function SpellFrame:RemoveTicksAfter(min)
	local indicators = self.indicators
	for i = #indicators,1,-1 do
		local ind = indicators[i]
		if ind.typeid == 'tick' and ind.time>min then
			self:Remove(i)
		end
	end
	self.nexttick = nil
end

function mainframe:CLEU_OtherInterestingSpell(time, event, srcguid,srcname,srcflags, destguid,destname,destflags, spellid,spellname)
	local now = GetTime()

	if EventHorizon.otherIDs[spellname] then
		local id = EventHorizon.otherIDs[spellname]
		if event == 'SPELL_CAST_SUCCESS' or event == 'SPELL_DAMAGE' and id.isGlyph then		-- Glyph refresh
			if id.last and (now < (id.last + 0.9)) then		-- Throttle heavily, don't want stacks getting blown
				--debug("Ignoring "..event.." from "..spellname.." at "..now)
				return
			else
				--debug("Interesting spell potentially not in frame detected! Spell: "..spellname,event)
				id.last = now
				local bar
				local frame = EventHorizon.barframes
				for i in pairs(frame) do
					local gr = frame[i].glyphrefresh or nil
					local gs = frame[i].glyphstacks or nil
					if gr and (gr[3] == spellname) then
						if gs[destguid] then
							gs[destguid] = gs[destguid] - 1
							frame[i].stacks:SetText(gs[destguid] > 0 and gs[destguid] or nil)
						end
						--debug("SUCCESS! "..gr[3].." has triggered "..frame[i].auraname)
					end
				end
			end
		end
	end
end

function SpellFrame:COMBAT_LOG_EVENT_UNFILTERED(time, event, srcguid,srcname,srcflags, destguid,destname,destflags, spellid,spellname)
	local now = GetTime()
	
	--debug(event, spellname, spellid)
	if event == 'SPELL_AURA_REMOVED' then
		if self.glyphrefresh then
			self.glyphstacks[destguid] = 0
		end
	end
	
	if event == 'SPELL_CAST_SUCCESS' then
		--debug('SPELL_CAST_SUCCESS',destguid)
		self.castsuccess[destguid] = now

		if self.glyphrefresh then
			if EventHorizon.glyphs[self.glyphrefresh[2]] then
				self.glyphstacks[destguid] = self.glyphrefresh[1]
				self.stacks:SetText(self.glyphstacks[destguid])
			end
		end
	elseif tickevents[event] then
		if UnitGUID(self.auraunit or 'target')==destguid then
			local tick = self:AddIndicator('tick', 'tick', now)
			tick.happened = true
			
			if (self.dot and (self.stop and self.stop ~= nil)) then
				if self.isHasted and self.ticks then
					self.dotMod = now - self.ticks.last
					self.ticks.last = now
				end
				self:RemoveTicksAfter(now) -- Reconstruct ticks from spellframe info
				local nexttick = now+(self.dotMod or self.dot)
				self.nexttick = nil
				while nexttick<=(self.stop+0.2) do -- Account for lag
					if now+future<nexttick then
						-- The next tick is not visible yet.
						self.nexttick = nexttick
						self.lasttick = self.stop
						break
					end
					if now+past<=nexttick then
						-- The next tick is visible.
						local tick = self:AddIndicator('tick', 'tick', nexttick)
						if nexttick<=now then
							tick.happened = true
						end
						self.latesttick = nexttick
					end
					nexttick=nexttick+(self.dotMod or self.dot)
				end	
			end
		end
	end
end

function SpellFrame:UNIT_AURA_refreshable(unitid)
	if unitid~=self.auraunit then return end
	local name, rank, icon, count, debuffType, duration, expirationTime, source = self.AuraFunction(self.auraunit, self.auraname)
	local afflicted = name and (self.unique or source=='player') and (not self.minstacks or count>=self.minstacks)
	local addnew, refresh
	local now = GetTime()
	local guid = UnitGUID(self.auraunit or 'target')
	-- First find out if the debuff was refreshed.
	local start
	
	if self.textureID and self.textureID ~= icon then
		return
	end
	
	if self.aurasegment and expirationTime == 0 and duration == 0 then	-- Timeless aura, bar exists (Overkill)
		for i = #self.indicators,1,-1 do
			self:Remove(i)
		end
		self.aurasegment = nil
		self.nexttick = nil
		self.stacks:SetText()
		return
	end
	
	if expirationTime == 0 then
		return
	end
	
	if afflicted then
		start = expirationTime-duration
		if self.targetdebuff then
			if self.targetdebuff.stop == expirationTime then
				start = self.targetdebuff.start
			else
				-- Check for refresh. 
				if start < self.targetdebuff.stop then
					local totalduration = self.targetdebuff.stop - self.targetdebuff.start
					local lasttick = self.targetdebuff.stop - math.fmod(totalduration, self.dotMod or self.dot)
					local success = self.castsuccess[guid]
					local not_recast = true -- Poisons are never actually recast, so we default to true here, because success will always be nil.
					if success then
						not_recast = math.abs(success-start)>0.5
					end
					if not_recast and start < lasttick then
						-- The current debuff was refreshed.
						start = self.targetdebuff.start
						refresh = true
					end
				end
			end
		end
		if self.aurasegment then
			if expirationTime~=self.aurasegment.stop and not refresh then
				-- The current debuff was replaced.
				self.aurasegment.stop = start-0.2
				self:RemoveTicksAfter(start)

				--debug('replaced')
				addnew = true
			end
		else
			addnew = true
		end
		if count>0 then
			self.stacks:SetFormattedText('%d',count)
		end
	else
		if self.aurasegment then
			if math.abs(self.aurasegment.stop - now)>0.3 then
				-- The current debuff ended.
				self.aurasegment.stop = now
			end
			self:RemoveTicksAfter(now)
			self.aurasegment = nil
			self.targetdebuff = nil
			self.stacks:SetText()
		end
	end
	self:UpdateDoT(addnew, source, now, start, expirationTime, name, refresh, guid)
end

function SpellFrame:UpdateDoT(addnew, source, now, start, expirationTime, name, refresh, guid)
	local addticks
	local isHasted
	local checkDoT = self.auranamePrimary or name
	local isPrimary = checkDoT == name or nil
	
	local targ = UnitName(self.auraunit)
	if addnew then
		--debug('addnew', start, expirationTime)
		local typeid = (source=='player' and self.isType) or (self.isType == 'debuffmine' and source~='player' and 'debuff')
		if self.cast and self.useSmalldebuff then
			self.aurasegment = self:AddSegment(typeid, 'smalldebuff', start, expirationTime)
			local hastedcasttime = select(7, GetSpellInfo(self.spellname))/1000
			self.cantcast = self:AddSegment(typeid, 'cantcast', start, expirationTime-hastedcasttime)
			self.aurasegment.lastunit = targ
		else
			self.aurasegment = self:AddSegment(typeid, 'default', start, expirationTime)
			self.aurasegment.lastunit = targ
		end
		-- Add visible ticks.
		if self.dot and isPrimary then
			addticks = start
		end
		if self.debuffs then
			-- Refreshable only.
			self.targetdebuff = {start=start, stop=expirationTime}
			self.debuffs[guid] = self.targetdebuff
		end
	elseif refresh then
		-- debug('refresh', start, expirationTime)
		-- Note: refresh requires afflicted and self.targetdebuff. Also, afflicted and not self.debuff implies addnew.
		-- So we can get here only if afflicted and self.debuff and self.targetdebuff.
		self.aurasegment.stop = expirationTime
		self.targetdebuff.stop = expirationTime
		if self.latesttick then
			addticks = self.latesttick
		end
	end
	if addticks then
		local nexttick = addticks+(self.dotMod or self.dot)
		self.nexttick = nil
		
		if self.hasted then
			if type(self.hasted) == 'number' then
				for glyph,i in pairs(EventHorizon.glyphs) do
					if glyph == self.hasted then isHasted = true end
				end
			elseif type(self.hasted) == 'table' then
				isHasted = (select(5,GetTalentInfo(self.hasted[1],self.hasted[2]))) > 0 or nil
			end
		end
		
		if isHasted and self.ticks then
				local tooShort = self.ticks.last and (now - self.ticks.last < 1.5) and true or nil
				self.ticks.last = self.ticks.last and ((now - self.ticks.last < 3) and not tooShort) and self.ticks.last or now
				self.isHasted = true
		elseif isHasted and self.expectedTicks then
				self.dotMod = (expirationTime - start)/self.expectedTicks
		end
		
		while nexttick<=expirationTime do
			if now+future<nexttick then
				-- The next tick is not visible yet.
				self.nexttick = nexttick
				self.lasttick = expirationTime
				break
			end
			if now+past<=nexttick then
				-- The next tick is visible.
				local tick = self:AddIndicator('tick', 'tick', nexttick)
				if nexttick<=now then
					tick.happened = true
				end
				self.latesttick = nexttick
			end
			nexttick=nexttick+(self.dotMod or self.dot)
		end
	end
	self.start, self.stop = start, expirationTime
end

function SpellFrame:PLAYER_REGEN_ENABLED()
	local thresh = GetTime() - 10
	local remove = {}
	for guid,data in pairs(self.debuffs) do
		if data.stop < thresh then
			tinsert(remove, guid)
		end
	end
	for _,guid in ipairs(remove) do
		--debug('removing',guid,self.spellname)
		self.debuffs[guid]=nil
	end
end

function SpellFrame:PLAYER_TARGET_CHANGED_refreshable()
	--debug(self.spellname, 'PLAYER_TARGET_CHANGED_refreshable')
	if self.auraunit ~= 'target' then return end
	if self.aurasegment then
		--debug(self.spellname, 'removing old')
		for i = #self.indicators,1,-1 do
			local ind = self.indicators[i]
			if typeids[ind.typeid] then
				self:Remove(i)
			end
		end
		self.aurasegment = nil
		self.targetdebuff = nil
		self.nexttick = nil
		self.stacks:SetText()
	end

	if UnitExists('target') then
		if UnitIsDead('target') then
			self.debuffs[UnitGUID('target')] = nil
		else
			self.targetdebuff = self.debuffs[UnitGUID('target')]
			--if self.targetdebuff then debug(self.spellname, 'have old') end
			self:UNIT_AURA('target')
			--if self.aurasegment then debug(self.spellname, 'added new') end
			if self.glyphstacks and self.aurasegment then
				if self.glyphstacks[UnitGUID('target')] and (self.glyphstacks[UnitGUID('target')] > 0) then
					self.stacks:SetText(self.glyphstacks[UnitGUID('target')])
				end
			end
		end
	end
end

function SpellFrame:SPELL_UPDATE_COOLDOWN()
	local start, duration, enabled = GetSpellCooldown(self.spellname)
	local ready = enabled==1 and start~=0 and duration and start+duration
	if ready and duration>1.5 then
		-- The spell is on cooldown, but not just because of the GCD.
		if self.cooldown ~= ready then
			--if self.coolingdown and (self.coolingdown.stop>start or self.coolingdown.start<ready) then debug('overlapping cooldown', self.spellname, start, ready) end
			self.coolingdown = self:AddSegment('cooldown', 'cooldown', start, ready)
			self.cooldown = ready
		end
	else
		if self.coolingdown then
			-- The spell was on cooldown, but now it's not (or it's on cooldown, but just because of the GCD).
			local now = GetTime()
			-- See when the cooldown is ready. If the spell is currently on GCD, check the GCD end; otherwise check now.
			if self.cooldown > (ready or now) then
				-- The cooldown ended earlier.
				self.coolingdown.stop = now
			end
			self.coolingdown = nil
		end
		self.cooldown = nil
	end
end

-- A SpellFrame is active (i.e. listening to events) iff the talent requirements are met.
-- The table activeFrames contains all the active frames.
-- If the stance requirement is not met, the frame is hidden, but still active.
function SpellFrame:Deactivate()
	if not self.isActive then return end
	--debug('unregistering events for', self.spellname)
	self:UnregisterAllEvents()
	if self.interestingCLEU then
		mainframe.framebyspell[self.spellname] = nil
	end
	self:Hide()
	for index=#self.indicators,1,-1 do
		self:Remove(index)
	end
	self.isActive = nil
end

function SpellFrame:Activate()
	if self.isActive then return end
	--debug('registering events for', self.spellname)
	for event in pairs(self.interestingEvent) do
		self:RegisterEvent(event)
	end
	if self.interestingCLEU then
		mainframe.framebyspell[self.spellname] = self
	end

	self:Show()
	self.isActive = true
end

function mainframe:PLAYER_TALENT_UPDATE()
	--debug('PLAYER_TALENT_UPDATE')
	EventHorizon:CheckTalents()
end

do
	local function HaveRequiredTalent(config)
		local result = true
		if config.requiredTalent then
			-- If requiredPoints is present, currentRank==requiredPoints is required.
			-- Otherwise, currentRank>0 is required.
			for i,req in ipairs(config.requiredTalent) do
				local tab,index,requiredPoints = unpack(req)
				local currentRank = (select(5,GetTalentInfo(tab,index))) or 0
				--debug(tab,index,requiredPoints or 'nil',currentRank)
				--if (requiredPoints and currentRank~=requiredPoints) or (not requiredPoints and currentRank==0) then
					-- TODO Create the frame anyways and register for talent changes.
				--	return
				--end
				if requiredPoints then
					result = currentRank == requiredPoints
				else
					result = currentRank > 0
				end
				if not result then
					break
				end
			end
		end
		return result
	end
	
	local function HaveRequiredGlyph(config)
		local result = true
		if config.requiredGlyph then
			result = nil
			for i = 1, 6 do
				local enabled,_,gsID,_ = GetGlyphSocketInfo(i)
				if enabled and gsID then
					if gsID == config.requiredGlyph then
						result = true
						--debug("Required glyph detected: ",gsID,EventHorizon.glyphs[gsID])
					end
				end
			end
		end
		return result
	end
	

	function EventHorizon:CheckTalents()
		table.wipe(activeFrames)
		for i,config in ipairs(self.barconfigs) do
			local frameNeeded = HaveRequiredTalent(config) and HaveRequiredGlyph(config) or nil
			-- Check if there already is a frame
			local spellframe = self.barframes[i]
			local frameExists = spellframe~=nil
			--debug('checking frame',i,': frameExists',frameExists,', frameNeeded ',frameNeeded)

			if frameNeeded then
				if frameExists then
					spellframe:Activate()
				else
					spellframe = self:CreateSpellBar(config)
					self.barframes[i] = spellframe
				end
				table.insert(activeFrames, spellframe)
			else
				if frameExists then
					spellframe:Deactivate()
				end
			end
		end

		local activate = #activeFrames > 0
		self:Activate(activate)
		if activate then
			mainframe:UPDATE_SHAPESHIFT_FORM()
		end

		--[[
		for i,f in ipairs(activeFrames)do
			debug('active frame',i,': ',f.spellname)
		end
		--]]
	end
end

function mainframe:UPDATE_SHAPESHIFT_FORM()
	--debug('UPDATE_SHAPESHIFT_FORM', 'GetShapeshiftForm()',GetShapeshiftForm(),'GetShapeshiftForm(true)',GetShapeshiftForm(true), 'GetNumShapeshiftForms()',GetNumShapeshiftForms())
	local stance = GetShapeshiftForm()
	-- On PLAYER_LOGIN, GetShapeshiftForm() sometimes returns a bogus value (2 on a priest with 1 form). No need to act upon it.
	if not stance or (GetNumShapeshiftForms() and stance>GetNumShapeshiftForms()) then return end
	mainframe:SetHeight(1)
	local numframes = 0
	
	EventHorizon.glyphs = {}
	for i = 1, 6 do
		local enabled,_,gsID,_ = GetGlyphSocketInfo(i)
		if enabled and gsID then
			EventHorizon.glyphs[gsID] = true
			--debug("Glyph detected: ",gsID,EventHorizon.glyphs[gsID])
		end
	end
	for i,spellframe in ipairs(activeFrames) do
		local shown = spellframe:IsShown()
		if spellframe.stance == stance and not shown then
			-- Show
			spellframe:Show()
			shown = true
		elseif spellframe.stance and spellframe.stance ~= stance and shown then
			-- Hide
			spellframe:Hide()
			for i,indicator in ipairs(spellframe.indicators) do
				indicator:Hide()
			end
			shown = false
		end

		if shown then
			numframes = numframes+1
			spellframe.icon:ClearAllPoints()
			spellframe.icon:SetPoint('TOPLEFT', mainframe, 'TOPLEFT', 0, -(numframes-1) * (barheight+barspacing))
		end
	end
	if numframes>0 then
		mainframe:SetHeight(numframes * (barheight+barspacing) - barspacing)
		mainframe:Show()
	else
		mainframe:Hide()
	end
	
	return true
end
mainframe.UPDATE_SHAPESHIFT_FORMS = mainframe.UPDATE_SHAPESHIFT_FORM
mainframe.GLYPH_ADDED = mainframe.UPDATE_SHAPESHIFT_FORM
mainframe.GLYPH_REMOVED = mainframe.UPDATE_SHAPESHIFT_FORM
mainframe.GLYPH_DISABLED = mainframe.UPDATE_SHAPESHIFT_FORM
mainframe.GLYPH_ENABLED = mainframe.UPDATE_SHAPESHIFT_FORM
mainframe.GLYPH_UPDATED = mainframe.UPDATE_SHAPESHIFT_FORM

-- GCD indicator
local gcdIndicator, gcdSpellName
do
	local gcdend

	local function OnUpdate(frame, elapsed)
		if gcdend then
			local now = GetTime()
			if gcdend<=now then
				gcdend = nil
				gcdIndicator:Hide()
			else
				local diff = now+past
				local p = (gcdend-diff)*scale
				if p<=1 then
					gcdIndicator:SetPoint('RIGHT', frame, 'RIGHT', (p-1)*barwidth+onepixelwide, 0)
					gcdIndicator:Show()
				end
			end
		end
	end
	function mainframe:SPELL_UPDATE_COOLDOWN()
		if gcdIndicator then
			local start, duration = GetSpellCooldown(gcdSpellName)
			if start and duration and duration>0 then
				gcdend = start+duration
				mainframe:SetScript('OnUpdate', OnUpdate)
			else
				gcdend = nil
				gcdIndicator:Hide()
				mainframe:SetScript('OnUpdate', nil)
			end
		end
	end
end


-- Dispatch the CLEU.
function mainframe:COMBAT_LOG_EVENT_UNFILTERED(time, event, srcguid,srcname,srcflags, destguid,destname,destflags, spellid,spellname)
	if srcguid~=playerguid or event:sub(1,5)~='SPELL' then return end
	local spellframe = self.framebyspell[spellname]
	if EventHorizon.otherIDs[spellname] then
		mainframe:CLEU_OtherInterestingSpell(time, event, srcguid,srcname,srcflags, destguid,destname,destflags, spellid,spellname)
	end
	if spellframe then
		if spellframe.interestingCLEU[event] then 
			spellframe:COMBAT_LOG_EVENT_UNFILTERED(time, event, srcguid,srcname,srcflags, destguid,destname,destflags, spellid,spellname) 
		end
	end
end

function EventHorizon:LoadClassModule()
	local class = select(2,UnitClass('player'))
	
	class = class:sub(1,1)..class:sub(2):lower() -- 'WARLOCK' -> 'Warlock'
	
	local name, title, notes, enabled, loadable, reason, security = GetAddOnInfo('EventHorizon_'..class)
	if not enabled or not loadable then 
		return 
	end
	local loaded, reason = LoadAddOn(name)
	if loaded and self.InitializeClass then
		return true
	end
end

--[[
spellid: number, rank doesn't matter
abbrev: string
config: table
{
	cast = <boolean>,
	channeled = <boolean>,
	numhits = <number of hits per channel>,
	cooldown = <boolean>,
	debuff = <boolean>,
	dot = <tick interval in s, requires debuff>,
	refreshable = <boolean>,
}
--]]
function EventHorizon:NewSpell(config)
	local spellid = config.spellID
	if type(spellid)~='number' then 
		return 
	end

	-- Make requiredTalent a list of required talents if only one is specified.
	if config.requiredTalent and type(config.requiredTalent[1])~='table' then
		config.requiredTalent = {config.requiredTalent}
	end

	table.insert(self.barconfigs, config)
end

--Set spellframe attributes separately from bar creation. Needed for custom bars that use aura/event info.
local function SetSpellAttributes(spellframe,config)
	local interestingEvent = {}
	local config = config
	local otherids = EventHorizon.otherIDs
	
	spellname = spellframe.spellname
	spellframe = EventHorizon[spellname]
	interestingEvent['UNIT_SPELLCAST_SENT'] = true
			
	if config.channeled then
		interestingEvent['UNIT_SPELLCAST_CHANNEL_START'] = true
		interestingEvent['UNIT_SPELLCAST_CHANNEL_STOP'] = true
		interestingEvent['UNIT_SPELLCAST_CHANNEL_UPDATE'] = true
		if config.numhits then
			-- Register for the CLEU tick event.
			spellframe.interestingCLEU = {SPELL_DAMAGE=true}
			spellframe.numhits = config.numhits
		end
	elseif config.cast then
		spellframe.cast = config.cast
		interestingEvent['UNIT_SPELLCAST_START'] = true
		interestingEvent['UNIT_SPELLCAST_STOP'] = true
		interestingEvent['UNIT_SPELLCAST_DELAYED'] = true
		if ((config.debuff or config.playerbuff) and type(config.debuff or config.playerbuff) == 'boolean') then
			spellframe.useSmalldebuff = true
		end
	end

	if config.cooldown then
		spellframe.cooldown = config.cooldown
		interestingEvent['SPELL_UPDATE_COOLDOWN'] = true
	end

	if config.stance then
		spellframe.stance = config.stance
	end

	if config.debuff then
		spellframe.isType = 'debuffmine'
		spellframe.AuraFunction = UnitDebuff
		spellframe.auraunit = config.auraunit or 'target'
		if type(config.debuff)=='number' then
			spellframe.auraname = (GetSpellInfo(config.debuff))
		elseif type(config.debuff)=='table' then
			spellframe.auranamePrimary = (GetSpellInfo(config.spellID))
			spellframe.auraname = {}
			for i,id in ipairs(config.debuff) do
				tinsert(spellframe.auraname, (GetSpellInfo(id)))
			end
			spellframe.AuraFunction = UnitDebuffUnique
		else
			spellframe.auraname = spellname
		end
		interestingEvent['UNIT_AURA'] = true
		interestingEvent['PLAYER_TARGET_CHANGED'] = true
		spellframe.minstacks = config.minstacks
		if config.dot then
			spellframe.dot = config.dot
			spellframe.interestingCLEU = {SPELL_PERIODIC_DAMAGE=true}
			if config.refreshable then
				spellframe.UNIT_AURA = SpellFrame.UNIT_AURA_refreshable
				spellframe.PLAYER_TARGET_CHANGED = SpellFrame.PLAYER_TARGET_CHANGED_refreshable
				interestingEvent['PLAYER_REGEN_ENABLED'] = true
				spellframe.interestingCLEU.SPELL_CAST_SUCCESS=true
				spellframe.debuffs = {}
				spellframe.castsuccess = {}
			end
			if config.hasted then
				spellframe.hasted = config.hasted
				if config.expectedTicks and type(config.expectedTicks) == "number" then
					spellframe.expectedTicks = config.expectedTicks
				else
					spellframe.ticks = {}
				end
			end
		end
	elseif config.playerbuff then
		spellframe.isType = 'playerbuff'
		spellframe.AuraFunction = UnitBuff
		spellframe.auraunit = config.auraunit or 'player'
		if config.auraunit then
			interestingEvent['PLAYER_TARGET_CHANGED'] = true
		end
		spellframe.alwaysrefresh = true
		interestingEvent['UNIT_AURA'] = true
		if type(config.playerbuff)=='number' then
			spellframe.auraname = (GetSpellInfo(config.playerbuff))
		elseif type(config.playerbuff)=='table' then
			spellframe.auraname = {}
			spellframe.auranamePrimary = (GetSpellInfo(config.spellID))
			for i,id in ipairs(config.playerbuff) do
				tinsert(spellframe.auraname, (GetSpellInfo(id)))
			end
			spellframe.AuraFunction = UnitBuffUnique
		else
			spellframe.auraname = spellname
		end
		--[[
		if type(config.playerbuff)=='number' then
			spellframe.auraname = (GetSpellInfo(config.playerbuff))
		else
			spellframe.auraname = spellname
		end
		]]--
		if config.internalcooldown then
			spellframe.internalcooldown = config.internalcooldown
		end
		if config.dot then
			spellframe.dot = config.dot
			spellframe.interestingCLEU = {SPELL_PERIODIC_HEAL=true}
			if config.refreshable then
				spellframe.UNIT_AURA = SpellFrame.UNIT_AURA_refreshable
				spellframe.PLAYER_TARGET_CHANGED = SpellFrame.PLAYER_TARGET_CHANGED_refreshable
				interestingEvent['PLAYER_REGEN_ENABLED'] = true
				spellframe.interestingCLEU.SPELL_CAST_SUCCESS=true
				spellframe.debuffs = {}
				spellframe.castsuccess = {}
			end
			if config.hasted then
				spellframe.hasted = config.hasted
				if config.expectedTicks and type(config.expectedTicks) == "number" then
					spellframe.expectedTicks = config.expectedTicks
				else
					spellframe.ticks = {}
				end
			end
		end
	end
	
	if config.cleu or config.event then -- Register custom CLEU events.
		if config.event then -- Optional alias for the forgetful
			config.cleu = config.event
		end
		if not spellframe.interestingCLEU then
			spellframe.interestingCLEU = {}
		end
		if type(config.cleu) == 'string' then -- Single event
			spellframe.interestingCLEU[config.cleu] = true
		elseif type(config.cleu) == 'table' then
			for i in pairs(config.cleu) do -- Multiple events
				spellframe.interestingCLEU[ config.cleu[i] ] = true
			end
		end
	end	
	
	if config.glyphrefresh and type(config.glyphrefresh) == "table" then
		spellframe.glyphrefresh = config.glyphrefresh
		spellframe.glyphstacks = {}
		if not otherids[config.glyphrefresh[3]] then
			otherids[config.glyphrefresh[3]] = {}
		end
		otherids[config.glyphrefresh[3]][spellname] = true
		otherids[config.glyphrefresh[3]].isGlyph = true
		spellframe.interestingCLEU["SPELL_AURA_REMOVED"] = true
	end
	
	if config.unique then
		spellframe.unique = config.unique
	end
	
	if config.textureID then
		spellframe.textureID = config.textureID
	end
	
	return interestingEvent
end

function EventHorizon:CreateSpellBar(config)
	local spellid = config.spellID
	local spellframe = CreateFrame('Frame', nil, mainframe)
	mainframe.numframes = mainframe.numframes+1
	mainframe:SetHeight(mainframe.numframes * (barheight+barspacing))

	local spellname, rank, tex = GetSpellInfo(spellid)
	--debug('creating frame for ',spellname)
	spellframe.spellname = spellname
	self[spellname] = spellframe

	-- Create the bar.
	spellframe.indicators = {}
	spellframe:SetBackdrop{bgFile = bartexture}
	spellframe:SetBackdropColor(unpack(barbgcolor))

	-- Create and set the spell icon.
	local icon = spellframe:CreateTexture(nil, 'BORDER')
	icon:SetTexture(tex)
	if not self.config.iconborder then
		local x=0.07
		local y=1-x
		icon:SetTexCoord(x,y,x,y)
	end
	spellframe.icon = icon

	local stacks = spellframe:CreateFontString(nil, 'OVERLAY')
	stacks:SetFontObject('NumberFontNormalSmall')
	spellframe.stacks = stacks

	local meta = getmetatable(spellframe)
	if meta and meta.__index then
		local metaindex = meta.__index
		setmetatable(spellframe, {__index = 
		function(self,k) 
			if SpellFrame[k] then 
				self[k]=SpellFrame[k] 
				return SpellFrame[k] 
			end 
			return metaindex[k]
		end})
	else
		setmetatable(spellframe, {__index = SpellFrame})
	end
	
	local interestingEvent = SetSpellAttributes(self[spellname],config)
	spellframe.interestingEvent = interestingEvent
	
	spellframe:SetScript('OnEvent', EventHandler)
	spellframe:SetScript('OnUpdate', spellframe.OnUpdate)

	-- Layout
	icon:SetPoint('TOPLEFT', mainframe, 'TOPLEFT', 0, -(mainframe.numframes-1) * (barheight+barspacing))
	spellframe:SetPoint('TOPLEFT', icon, 'TOPRIGHT')
	spellframe:SetWidth(barwidth)
	spellframe:SetHeight(barheight)
	icon:SetWidth(barheight)
	icon:SetHeight(barheight)
	stacks:SetPoint('BOTTOMRIGHT', icon, 'BOTTOMRIGHT')
	stacks:SetJustifyH('RIGHT')

	spellframe:Activate()

	return spellframe
end

function EventHorizon:Activate(...)
	local activate = select('#',...) == 0 or ...
	--debug('Activate',activate, ...)
	if not activate then
		return self:Deactivate()
	end

	if self.isActive then
		return
	end

	for k,v in pairs(reloadEvents) do
		mainframe:RegisterEvent(k)
	end	

	mainframe:Show()

	self.isActive = true
end

function EventHorizon:Deactivate()
	if self.isActive==false then
		return
	end
	mainframe:UnregisterAllEvents()
	mainframe:RegisterEvent('PLAYER_TALENT_UPDATE')

	mainframe:Hide()

	self.isActive = false
end

--[[
Should only be called after the DB is loaded and spell and talent information is available.
--]]
function EventHorizon:Initialize()
	-- Make sure this function is called only once.
	self.Initialize = nil

	--debug('GetTalentInfo(1,1)',GetTalentInfo(1,1))
	playerguid = UnitGUID('player')
	if not playerguid then
		error('no playerguid')
	end
	self.db = EventHorizonDB
	mainframe.spellframes = activeFrames

	-- Create the main and spell frames.
	mainframe:SetHeight(1)
	mainframe.numframes = 0
	mainframe.framebyspell= {}
	mainframe:SetScript('OnEvent', EventHandler)
	mainframe:SetScale(self.config.scale or 1)

	if not self:LoadClassModule() then
		return
	end

	self:ApplyConfig()

	self:InitializeClass()

	mainframe:SetWidth(barwidth + barheight)
	self:CheckTalents()

	if self.config.backdrop then
		self:SpawnStyleFrame()		-- Spawn backdrop frame.
	end
		
	-- Create the indicator for the current time.
	-- Bugfix: When the UI scale is at a very low setting, textures with a width of 1
	-- were not visible in some resolutions.
	local effectiveScale = mainframe:GetEffectiveScale()
	if effectiveScale then
		onepixelwide = 1/effectiveScale
	end
	local nowI = CreateFrame('Frame',nil,mainframe)
	nowI:SetFrameLevel(20)
	local nowIndicator = nowI:CreateTexture('EventHorizonNowIndicator', 'OVERLAY')
	
	local nowleft = -past/(future-past)*barwidth-0.5 + barheight
	nowIndicator:SetPoint('BOTTOM',mainframe,'BOTTOM')
	nowIndicator:SetPoint('TOPLEFT',mainframe,'TOPLEFT', nowleft, 0)
	nowIndicator:SetWidth(onepixelwide)
	nowIndicator:SetTexture(1,1,1,1)

	local anchor = self.config.anchor or {'TOPRIGHT', 'EventHorizonHandle', 'BOTTOMRIGHT'}
	if anchor[2]=='EventHorizonHandle' then
		-- Create the handle to reposition the frame.
		local handle = CreateFrame('Frame', 'EventHorizonHandle', mainframe)
		handle:SetFrameStrata('HIGH')
		handle:SetWidth(10)
		handle:SetHeight(5)
		handle:EnableMouse(true)
		handle:SetClampedToScreen(1)
		handle:RegisterForDrag('LeftButton')
		handle:SetScript('OnDragStart', function(handle, button) handle:StartMoving() end)
		handle:SetScript('OnDragStop', function(handle) 
			handle:StopMovingOrSizing()
			local a,b,c,d,e = handle:GetPoint(1)
			if type(b)=='frame' then
				b=b:GetName()
			end
			self.db.point = {a,b,c,d,e}
		end)
		handle:SetMovable(true)

		mainframe:SetPoint(unpack(anchor))
		handle:SetPoint(unpack(self.db.point))

		handle.tex = handle:CreateTexture(nil, 'BORDER')
		handle.tex:SetAllPoints()
		handle:SetScript('OnEnter',function(frame) frame.tex:SetTexture(1,1,1,1) end)
		handle:SetScript('OnLeave',function(frame) frame.tex:SetTexture(1,1,1,0.1) end)
		handle.tex:SetTexture(1,1,1,0.1)
	end

	gcdSpellName = self.config.gcdSpellID and (GetSpellInfo(self.config.gcdSpellID))
	if gcdSpellName and self.config.gcdStyle then
		-- Create the GCD indicator, register cooldown event.
		gcdIndicator = mainframe:CreateTexture('EventHorizonGCDIndicator', 'BORDER')
		local nowleft = -past/(future-past)*barwidth-0.5 + barheight
		gcdIndicator:SetPoint('BOTTOM',mainframe,'BOTTOM')
		gcdIndicator:SetPoint('TOP',mainframe,'TOP')
		gcdIndicator:Hide()

		if self.config.gcdStyle == 'line' then
			gcdIndicator:SetWidth(onepixelwide)
		else
			gcdIndicator:SetPoint('LEFT',mainframe,'LEFT', nowleft, 0)
		end

		local gcdColor = self.config.gcdColor or {.5,.5,.5,.3}
		gcdIndicator:SetTexture(unpack(gcdColor))
	end

	mainframe:SetPoint(unpack(anchor))

	-- Upgrade DB.
	if not self.db.version then
		self.db.version = 1
		self.db.isActive = true
	end

	SLASH_EVENTHORIZON1 = '/eventhorizon'
	SLASH_EVENTHORIZON2 = '/ehz'
	SlashCmdList['EVENTHORIZON'] = function(msg)
		if msg == 'config' then
			InterfaceOptionsFrame_OpenToCategory('EventHorizon')
		else
			if self.isActive then
				self:Deactivate()
				mainframe:UnregisterEvent('PLAYER_TALENT_UPDATE')
			else
				self:Activate()
				EventHorizon:CheckTalents()
			end
			self.db.isActive = self.isActive
		end
	end

	if not self.db.isActive then
		self:Deactivate()
		mainframe:UnregisterEvent('PLAYER_TALENT_UPDATE')
	end
	
	self.isReady = true
end

function EventHorizon:ApplyConfig()
	past = self.config.past or -3
	future = self.config.future or 9
	barheight = self.config.height or 18
	barwidth = self.config.width or 150
	barspacing = self.config.spacing or 0
	scale = 1/(future-past)
	bartexture = self.config.bartexture or 'Interface\\Addons\\EventHorizon\\Smooth'
	barbgcolor = self.config.barbgcolor or {1,1,1,0.1}
	texturedbars = self.config.texturedbars
	local texturealpha = self.config.texturealphamultiplier or 1
	classburn = self.config.classburn or 0.7
	classalpha = self.config.classalpha or 0.3
	self.config.castLine = self.config.castLine or nil
	
	local classcolors = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
	self.classcolor = Clone(classcolors[select(2,UnitClass('player'))])

	for colorid,color in pairs(self.colors) do
		if color[1] == true then
			if color[2] then
				self.colors[colorid] = {self.classcolor.r * color[2], self.classcolor.g * color[2], self.classcolor.b * color[2], color[3] or classalpha}		-- For really bad reasons, this took a very long time to get right...
			else
				self.colors[colorid] = {self.classcolor.r, self.classcolor.g, self.classcolor.b, classalpha}
			end
		end
	end
	
	if texturedbars then
		for colorid,color in pairs(colors) do
			if colorid~='default' and color[4] then
				color[4] = texturealpha*color[4]
			end
		end
	end

	local default = layouts.default
	for typeid,layout in pairs(layouts) do
		if typeid~='default' then
			for k,v in pairs(default) do
				if layout[k]==nil then
					layout[k] = v
				end
			end
		end
		local topperc, bottomperc = -layout.top[3]/barheight
		if layout.bottom[1] == 'TOP' then
			bottomperc = -layout.bottom[3]/barheight
		else
			bottomperc = 1-layout.bottom[3]/barheight
		end
		layout.texcoords = {0, 1, topperc, bottomperc}
	end
end

local glyphCheck = function ()
	if EventHorizon.isActive == true then
		EventHorizon:CheckTalents()
	end	
end
	
do
	local frame = CreateFrame('Frame')
	frame:SetScript('OnEvent', EventHandler)

	local talentcheck = GetTalentInfo(1,1)
	if talentcheck then
		-- Talent information is already available, which means that the interface was reloaded.
		frame:RegisterEvent('PLAYER_LOGIN')
		function frame:PLAYER_LOGIN()
			EventHorizon:Initialize()
			self:UnregisterEvent('PLAYER_LOGIN')
		end
	else
		frame:RegisterEvent('PLAYER_ALIVE')
		function frame:PLAYER_ALIVE()
			EventHorizon:Initialize()
			self:UnregisterEvent('PLAYER_ALIVE')
		end
	end
	
	local frame2 = CreateFrame('Frame')
	frame2:SetScript('OnEvent', EventHandler)
	for k,v in pairs(reloadEvents) do
		if (k ~= 'COMBAT_LOG_EVENT_UNFILTERED') and (k ~= 'SPELL_UPDATE_COOLDOWN') then
			frame2:RegisterEvent(k)
			frame2[k] = glyphCheck
		end
	end
end

function EventHorizon:SpawnStyleFrame()
	self.SpawnStyleFrame = nil
	local stylebg = self.config.bg or 'Interface\\ChatFrame\\ChatFrameBackground'
	local styleborder = self.config.border or 'Interface\\Tooltips\\UI-Tooltip-Border'
	local stylebgcolor = self.config.bgcolor or {0,0,0,0.5}
	local stylebordercolor = self.config.bordercolor or {1,1,1,1}
	local styleinset = self.config.inset or {top = 2, bottom = 2, left = 2, right = 2}
	local stylepadding = self.config.padding or 3
	local styleedge = self.config.edgesize or 8

	local styleframe = CreateFrame('Frame',nil,mainframe)
	styleframe:SetFrameStrata('BACKGROUND')
	styleframe:SetPoint('TOPRIGHT', mainframe, 'TOPRIGHT', stylepadding, stylepadding)
	styleframe:SetPoint('BOTTOMLEFT', mainframe, 'BOTTOMLEFT', -stylepadding, -stylepadding)
	styleframe:SetBackdrop({
		bgFile = stylebg,
		edgeFile = styleborder, tileSize = 0, edgeSize = styleedge,
		insets = styleinset,
	})
	styleframe:SetBackdropColor(unpack(stylebgcolor))
	styleframe:SetBackdropBorderColor(unpack(stylebordercolor))
end
