local config = EventHorizon.config
local _,class = UnitClass('player')
local DK = class == "DEATHKNIGHT"
local Druid = class == "DRUID"
local Hunter = class == "HUNTER"
local Mage = class == "MAGE"
local Paladin = class == "PALADIN"
local Priest = class == "PRIEST"
local Rogue = class == "ROGUE"
local Shaman = class == "SHAMAN"
local Warlock = class == "WARLOCK"
local Warrior = class == "WARRIOR"

-- Past and future determine the number of seconds displayed before and after the current time.
config.past = -3
config.future = 12

-- Width and height of a single bar.
config.width = 150
config.height = 18

-- The space between two bars.
config.spacing = 0

-- The overall scale of the main frame.
config.scale = 1

-- By default, the main frame is parented to the handle to make it draggable.
-- The top-right of the frame is anchored to the top-left of the
-- handle, which makes the frame grow downwards. 
config.anchor = {"TOPRIGHT", "EventHorizonHandle", "BOTTOMRIGHT"}

-- Use the following line to anchor the bottom left of the frame to the top left of the handle, thus allowing the 
-- bars to grow upwards. You will only notice a difference on Druids, where bars are shown/hidden depending on stance.
--config.anchor = {"BOTTOMLEFT", "EventHorizonHandle", "TOPLEFT"}

-- Alternatively, set a fixed position for the main frame. The handle won't be created if the main frame is not going to be parented to it.
-- The following line will anchor the top of the main frame to the center of the screen, making the frame grow downwards.
--config.anchor = {"TOP","UIParent","CENTER",100,100}

-- When you anchor the bottom, the frame will grow upwards:
--config.anchor = {"BOTTOM","UIParent","CENTER",100,100}

-- Comment out or remove the following line if you want to keep bars with a solid color.
config.texturedbars = true
-- Most textures appear darker than a solid color because of the alpha value. To counteract this, 
-- the alpha value of the color is multiplied by the following value before it's applied to a texture.
config.texturealphamultiplier = 2

-- Create the backdrop frame and border.
config.backdrop = true
-- Set it to nil to remove it, like so:
--config.backdrop = nil

-- Use the default Blizzard icon border.
config.iconborder = false

-- Add full-frame casting lines for spellcasts over 1.5 seconds.
config.castLine = true

-- Display the end of the GCD as a thin line.
config.gcdStyle = 'line'
-- Alternatively, display the GCD as bar from now to the end.
--config.gcdStyle = 'bar'
-- To disable the GCD indicator, clear the style.
--config.gcdStyle = false
-- Set the color of the GCD indicator.
config.gcdColor = {1,1,1,0.5}

-- Backdrop, border, and texture settings. These are the defaults, tweak to taste.
-- Individual bar background.
config.bartexture = "Interface\\Addons\\EventHorizon\\Smooth"
-- Individual bar background color.
config.barbgcolor = {1,1,1,0}

-- Frame background.
config.bg = "Interface\\ChatFrame\\ChatFrameBackground"
-- Border texture.
config.border = "Interface\\Tooltips\\UI-Tooltip-Border"
-- Frame background color.
config.bgcolor = {0,0,0,0.6}
-- Border color.
config.bordercolor = {1,1,1,1}

-- Advanced settings.
-- Use this to shrink or enlarge the background. Negative numbers move the background edges further from its center, positive numbers move it closer.
config.inset = {top = 2, bottom = 2, left = 2, right = 2}
-- Extra space between the EventHorizon bars and the border.
config.padding = 3
-- Thickness of the frame's border.
config.edgesize = 8

-- Color of individual bar segments. Format is R,G,B,Alpha with a max of 1 and a minimum of 0. Math and other references may be done within these tables for finer control. Example: {125/255,125/255,240/255,1} or {0.15*1.38, 3/16, 0.1+c.debuff[3], 0.5}
-- Class coloring has two options, burn (darkness) and alpha (opacity). Burn has no max, but be careful how high the value is adjusted or you'll end up with pure white. Alpha is 0 to 1.
-- To color a bar or indicator by class, use {true, <burn>, <alpha>}. Note that the alpha portion is affected by config.texturealphamultiplier as with manual coloring.
local c = EventHorizon.colors
c.sent = {true,Priest and 0.7 or 1,1}			-- Marker line when a spellcast is sent to the server. Default = {true,1,1} (class colored, bright and opaque)
c.tick = {true,Priest and 0.7 or 1,1}			-- Tick markers. Default = {true,1,1} (class colored, bright and opaque)
c.casting = {0,1,0,0.3}							-- Casting bars. Default = {0,1,0,0.3} (green, 0.3 unmodified alpha)
c.cooldown = {1,1,1,0.3}						-- Cooldown bars. Default = {1,1,1,0.3} (white, 0.3 unmodified alpha)
c.debuffmine = {true,Priest and 0.7 or 1,0.3}	-- YOUR debuff bars. Default = {true,1,0.3} (class colored, bright, 0.3 unmodified alpha)
c.debuff = {true,0.5,0.3}						-- OTHER PLAYERS' debuff bars. Default = {true,0.5,0.3} (class colored, darkened by 50%, 0.3 unmodified alpha)
c.playerbuff = {true,Priest and 0.7 or 1,0.3}	-- Buff bars. Default = {true,1,0.3} (class colored, bright, 0.3 unmodified alpha)