EventHorizon_Vitals = {}
EventHorizon_Vitals.config = {}
EventHorizon_Vitals.backdrop = {}
EventHorizon_Vitals.colors = {}

local function printhelp(...)
	if select('#',...)>0 then
		return tostring((...)), printhelp(select(2,...))
	end
end

local function print(...)
	ChatFrame1:AddMessage('EventHorizon_Vitals: '..strjoin(',',printhelp(...)))
end

local layout
local _, class = UnitClass("player");
if ((class == "ROGUE")or(class == "DRUID")) then
	layout = "cp"
elseif class == "DEATHKNIGHT" then
	layout = "dk"
else
	layout = nil
end

local gtg = nil
local useEH = nil
local EventHandler
	
local time = GetTime()
local TTL, tTTL, hTTL = 0.05, 0.10, 3
local TSLU, tTSLU, hTSLU = 0, 0, 0

local rounding = 60000

local WSID = 32535
local WSID2 = 30161

local frame = CreateFrame('Frame','EventHorizon_VitalsFrame',UIParent)
local text = frame:CreateFontString(nil,'OVERLAY')
local combo = frame:CreateFontString(nil,'OVERLAY')
local mhT = frame:CreateFontString(nil,'OVERLAY')
local ohT = frame:CreateFontString(nil,'OVERLAY')
local sep = frame:CreateFontString(nil,'OVERLAY')

local r = {}
for i=1,6 do
	r[i] = frame:CreateFontString(nil,'OVERLAY')
end

local function EventHandler(self, event, ...)
	local f = self[event]
	if f then 
		f(self,...) 
	end 
end

local function Clone(t)
	local new = {}
	local i, v = next(t, nil)	-- i is an index of t, v = t[i]
	while i do
		new[i] = v
		i, v = next(t, i)
	end
	return new
end

local function Round(num,idp)
  if idp and idp>0 then
    local mult = 10^idp
    return math.floor(num * mult + 0.5) / mult
  end
  return math.floor(num + 0.5)
end

function EventHorizon_Vitals:CheckEH()
	if IsAddOnLoaded("EventHorizon") then
		useEH = true
	else
		useEH = nil
	end
	
end

function EventHorizon_Vitals:IsComboUser()		--Thanks to MalygosCP for the combo-vehicle check.
	local guid = "vehicle" and UnitGUID("vehicle")
	if guid then
		if((WSID==tonumber(strsub(guid, -12,-7),16)) or (WSID2==tonumber(strsub(guid, -12,-7),16))) then 
			return true
        else
			return nil
		end
	else
		if ((class == "ROGUE")or(class == "DRUID")) then
			return true
		else
			return nil
		end
	end
end

function EventHorizon_Vitals:EnterVehicle(unit)
	if unit == "player" then
		if (CanExitVehicle() and UnitPower("vehicle")) then
			local p = UnitPower("vehicle")
			local pt, pts = UnitPowerType("vehicle")
			if p > 1 and pt >= 0 then
				layout = "veh"
			else
				if ((class == "ROGUE")or(class == "DRUID")) then
					layout = "cp"
				elseif class == "DEATHKNIGHT" then
					layout = "dk"
				else
					layout = nil
				end
			end
		end
	else return end
end

function EventHorizon_Vitals:ExitVehicle(unit)
	if ((unit == "player") and (layout == "veh")) then
		if ((class == "ROGUE")or(class == "DRUID")) then
			layout = "cp"
		elseif class == "DEATHKNIGHT" then
			layout = "dk"
		else
			layout = nil
		end
	else return end
end

function EventHorizon_Vitals:UpdateCP()
	local c
	local u
	
	if layout == "cp" then
		u = "player"
	elseif layout == "veh" then
		u = "vehicle"
	end
	
	if EventHorizon_Vitals:IsComboUser() then
		c = GetComboPoints(u)
		if((pt == 3)or(c > 0)) then
			combo:SetText(c)
			combo:SetVertexColor(unpack(EventHorizon_Vitals.color["c"..c]))
		else
			combo:SetText("")
			combo:SetVertexColor(unpack(EventHorizon_Vitals.color["c"..c]))
		end
	else
		combo:SetText("")
	end
end

function EventHorizon_Vitals:UpdateRune(runeid,usable)
	local start,duration,ready = GetRuneCooldown(runeid)
	local rC = EventHorizon_Vitals.config.runeCountdown or 0
	
	if runeid > 6 then return else
		if ready == true then	
			r[runeid]:SetText(EventHorizon_Vitals.config.runeReady)
		else
			local cd = start + duration - GetTime()
			cd = Round(cd)
			if ((rC > 0) and (cd <= rC)) then
				r[runeid]:SetText(cd)
			else
				r[runeid]:SetText(EventHorizon_Vitals.config.runeCooling)
			end			
		end
	end
end

function EventHorizon_Vitals:UpdateRuneType(runeid)
	local rt = GetRuneType(runeid)
	local rc = EventHorizon_Vitals.color
	if runeid > 6 then return else
		r[runeid]:SetVertexColor(unpack(rc["r"..rt]))
	end
	EventHorizon_Vitals:UpdateRune(runeid)
end	

function EventHorizon_Vitals:SpawnStyleFrame()
	local st = EventHorizon_Vitals.backdrop
	local sf = CreateFrame('Frame','EventHorizon_VitalsBG',EventHorizon_VitalsFrame)
	
	sf:SetFrameStrata('BACKGROUND')
	if st.backdrop then
		if useEH then
			local eh = EventHorizon.config
			if eh.backdrop == true then
				sf:SetPoint('TOPRIGHT', EventHorizon_VitalsFrame, 'TOPRIGHT', eh.padding, eh.padding)
				sf:SetPoint('BOTTOMLEFT', EventHorizon_VitalsFrame, 'BOTTOMLEFT', -eh.padding, -eh.padding)
				sf:SetBackdrop({
					bgFile = eh.bg,
					edgeFile = eh.border, tileSize = 0, edgeSize = eh.edgesize,
					insets = eh.inset,
				})
				sf:SetBackdropColor(unpack(eh.bgcolor))
				sf:SetBackdropBorderColor(unpack(eh.bordercolor))
			else return end
		else
			sf:SetPoint('TOPRIGHT', EventHorizon_VitalsFrame, 'TOPRIGHT', st.padding, st.padding)
			sf:SetPoint('BOTTOMLEFT', EventHorizon_VitalsFrame, 'BOTTOMLEFT', -st.padding, -st.padding)
			sf:SetBackdrop({
				bgFile = st.bgFile,
				edgeFile = st.borderFile, tileSize = 0, edgeSize = st.edgesize,
				insets = st.inset,
			})
			sf:SetBackdropColor(unpack(st.bgColor))
			sf:SetBackdropBorderColor(unpack(st.borderColor))
		end
	end	
end

function EventHorizon_Vitals:ApplyFonts(fs,j,flag)		-- Ex: ApplyFonts(fontstring1,"LEFT")
	if flag == 1 then
		fs:SetFont(EventHorizon_Vitals.config.runeFont,EventHorizon_Vitals.config.runeFontSize)
	elseif flag == 2 then
		fs:SetFont(EventHorizon_Vitals.config.font,EventHorizon_Vitals.config.enchantFontSize)
	else
		fs:SetFont(EventHorizon_Vitals.config.font,EventHorizon_Vitals.config.fontSize)
	end

	fs:SetJustifyH(j)
	fs:SetHeight(EventHorizon_Vitals.config.height)
	fs:Show()
end

function EventHorizon_Vitals:ApplyConfig()
	if useEH then
		local eh = EventHorizon.config
	end
	local cc = EventHorizon_Vitals.config
	local offset = cc.offset
	
	classburn = self.config.classburn or 1
	classalpha = self.config.classalpha or 1

	local classcolors = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
	self.classcolor = Clone(classcolors[select(2,UnitClass('player'))])

	for colorid,color in pairs(self.color) do
		if type(color) == 'table' and color[1] == true then
			if color[2] then
				self.color[colorid] = {self.classcolor.r * color[2], self.classcolor.g * color[2], self.classcolor.b * color[2], color[3] or classalpha}		-- For really bad reasons, this took a very long time to get right...
			else
				self.color[colorid] = {self.classcolor.r, self.classcolor.g, self.classcolor.b, classalpha}
			end
		end
	end
	
	frame:SetHeight(cc.height)
	
	if cc.position == 'top' then
		frame:SetPoint("BOTTOMLEFT", EventHorizon.mainframe, "TOPLEFT", 0, offset)
		frame:SetPoint("BOTTOMRIGHT", EventHorizon.mainframe, "TOPRIGHT", 0, offset)
	elseif cc.position == 'bottom' then
		frame:SetPoint("TOPLEFT", EventHorizon.mainframe, "BOTTOMLEFT", 0, -offset)
		frame:SetPoint("TOPRIGHT", EventHorizon.mainframe, "BOTTOMRIGHT", 0, -offset)
	else
		frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
		print("WARNING! You have an invalid position set in your config or myconfig.lua file!")
	end
	
	local toffx = cc.textOffset[1]
	local toffy = cc.textOffset[2]
	
	ohT:SetPoint("RIGHT",frame,"RIGHT",-toffx,toffy)
	sep:SetPoint("RIGHT",ohT,"LEFT",-toffx,0)
	mhT:SetPoint("RIGHT",sep,"LEFT",-toffx,0)
	
	if ((layout == "cp") or (layout == "veh")) then
		combo:SetPoint("CENTER", EventHorizon_VitalsFrame, "CENTER", 0, toffy)
		text:SetPoint("LEFT", EventHorizon_VitalsFrame, "LEFT", toffx, toffy)
	elseif layout == "dk" then
		local rs = EventHorizon_Vitals.config.runeSpacing
		text:SetPoint("LEFT", EventHorizon_VitalsFrame, "LEFT", toffx, toffy)
		combo:Hide()
		for i=1,6 do
			if i == 1 then
				r[1]:SetPoint("LEFT", EventHorizon_VitalsFrame, "CENTER", -cc.runeSpacing * 2.5, cc.runeOffset)
			else
				r[i]:SetPoint("LEFT", r[i-1], "LEFT", cc.runeSpacing, 0)
			end
		end
	else
		text:SetPoint("LEFT", EventHorizon_VitalsFrame, "LEFT", toffx, toffy)
		combo:Hide()
	end
end

function EventHorizon_Vitals:Init()	
	if layout == "dk" then
		combo:Hide()
		EventHorizon_Vitals:ApplyFonts(text,"RIGHT")		
		for i=1,6 do
			EventHorizon_Vitals:ApplyFonts(r[i],"LEFT",1)
			r[i]:Show()
		end
	else
		EventHorizon_Vitals:ApplyFonts(text,"LEFT")
		if layout == "cp" then
			EventHorizon_Vitals:ApplyFonts(combo,"LEFT")
			combo:Show()
			for i=1,6 do
				r[i]:Hide()
			end
		end
		if layout == "veh" then
			combo:Show()
			EventHorizon_Vitals:ApplyFonts(combo,"RIGHT")
			for i=1,6 do
				r[i]:Hide()
			end
		end
	end

	if layout == "cp" then
		EventHorizon_Vitals:UpdateCP()
	elseif layout == "dk" then
		for i=1,6 do
			EventHorizon_Vitals:UpdateRuneType(i)
		end
	elseif layout == "veh" then
		EventHorizon_Vitals:UpdateCP()
	end
	
	EventHorizon_Vitals:ApplyFonts(mhT,'RIGHT',2)
	EventHorizon_Vitals:ApplyFonts(ohT,'RIGHT',2)
	EventHorizon_Vitals:ApplyFonts(sep,'RIGHT',2)
	
	EventHorizon_Vitals:ApplyConfig()
	frame:Show()
end
--[[
function EventHorizon_Vitals:Update()
	local p,pm,pt,pts
	if layout == "veh" then
		p = UnitPower("vehicle")
		pm = UnitPowerMax("vehicle")
		pt, pts = UnitPowerType("vehicle")
	else
		p = UnitPower("player")
		pm = UnitPowerMax("player")
		pt, pts = UnitPowerType("player")
	end
	
	local c
	if EventHorizon_Vitals.color[pts] then
		c = EventHorizon_Vitals.color[pts]
	else
		c = EventHorizon_Vitals.color.ERROR
	end
	
	local m = EventHorizon_Vitals.color.fade
	--print(p,pm,pt,pts)
	
	if p and pt then
		if(((pt == 1) or (pt == 6)) and(p == 0)) then
			text:SetVertexColor(c[1]*m,c[2]*m,c[3]*m)
		else
			text:SetVertexColor(c[1],c[2],c[3])
		end
		if not((pt == 1)or(pt == 6)) then
			if p == pm then
				text:SetVertexColor(c[1]*m,c[2]*m,c[3]*m)
			else
				text:SetVertexColor(c[1],c[2],c[3])
			end
		end
	else
		text:SetVertexColor(1,1,1)
	end

if EventHorizon_Vitals.color[pts] then
	text:SetText(p)
else
	text:SetText("")
end

TSLU = 0
end
]]--
function EventHorizon_Vitals.OnUpdate(self,elapsed)		-- DON'T TOUCH ME!
	TSLU = TSLU + elapsed
	tTSLU = tTSLU + elapsed
	hTSLU = hTSLU + elapsed
		
		-- Regularly-throttled power check.
		while (TSLU > TTL) do
			if not gtg then
				EventHorizon_Vitals:CheckEH()
				return
			else
				local p,pm,pt,pts
				if layout == "veh" then
					p = UnitPower("vehicle")
					pm = UnitPowerMax("vehicle")
					pt, pts = UnitPowerType("vehicle")
				else
					p = UnitPower("player")
					pm = UnitPowerMax("player")
					pt, pts = UnitPowerType("player")
				end
				
				local c
				if EventHorizon_Vitals.color[pts] then
					c = EventHorizon_Vitals.color[pts]
				else
					c = EventHorizon_Vitals.color.ERROR
				end
				
				local m = EventHorizon_Vitals.color.fade
				--print(p,pm,pt,pts)
				
				if p and pt then
					if(((pt == 1) or (pt == 6)) and(p == 0)) then
						text:SetVertexColor(c[1]*m,c[2]*m,c[3]*m)
					else
						text:SetVertexColor(c[1],c[2],c[3])
					end
					if not((pt == 1)or(pt == 6)) then
						if p == pm then
							text:SetVertexColor(c[1]*m,c[2]*m,c[3]*m)
						else
							text:SetVertexColor(c[1],c[2],c[3])
						end
					end
				else
					text:SetVertexColor(1,1,1)
				end

			if EventHorizon_Vitals.color[pts] then
				text:SetText(p)
			else
				text:SetText("")
			end

			TSLU = 0
			TSLU = TSLU - TTL
		end
			
		-- Highly throttled updates for DK runes
		while (tTSLU > tTTL) do
			if layout == "dk" then
				local rC = EventHorizon_Vitals.config.runeCountdown or 0
				if (rC > 0) then
					for i=1,6 do
						local rt = GetRuneType(i)
						local rc = EventHorizon_Vitals.color
						local start,duration,ready = GetRuneCooldown(i)
						local rC = EventHorizon_Vitals.config.runeCountdown or 0
						r[i]:SetVertexColor(unpack(rc["r"..rt]))
						
						if ready == true then
							r[i]:SetText(EventHorizon_Vitals.config.runeReady)
						else
							local cd = start + duration - GetTime()
							cd = Round(cd)
							if ((rC > 0) and (cd <= rC)) then
								r[i]:SetText(cd)
							else
								r[i]:SetText(EventHorizon_Vitals.config.runeCooling)
							end
						end
					
						tTSLU = 0
					end
				end
			end
			tTSLU = tTSLU - tTTL
		end
		
		-- Weapon enchant updates, even more heavily throttled.
		while (hTSLU > hTTL) do
			local mh,mhEx,_,oh,ohEx,_ = GetWeaponEnchantInfo()
			--print(mh,mhEx,oh,ohEx)
			if mh then
				--print(r)
				local mt = Round(mhEx/rounding)
				mhT:SetText(mt)
				if mt <= 10 then
					mhT:SetVertexColor(unpack(EventHorizon_Vitals.color.weaponexpiring))
				elseif mt <= 30 then
					mhT:SetVertexColor(unpack(EventHorizon_Vitals.color.weaponhalf))
				else
					mhT:SetVertexColor(unpack(EventHorizon_Vitals.color.weaponfresh))
				end
			else
				mhT:SetText('')
				mhT:SetVertexColor(unpack(EventHorizon_Vitals.color.weaponexpiring))
			end
			if oh then
				local ot = Round(ohEx/rounding)
				ohT:SetText(ot)
				if ot <= 10 then
					ohT:SetVertexColor(unpack(EventHorizon_Vitals.color.weaponexpiring))
				elseif ot <= 30 then
					ohT:SetVertexColor(unpack(EventHorizon_Vitals.color.weaponhalf))
				else
					ohT:SetVertexColor(unpack(EventHorizon_Vitals.color.weaponfresh))
				end
			else
				ohT:SetText('')
				ohT:SetVertexColor(unpack(EventHorizon_Vitals.color.weaponexpiring))
			end
			
			if not (mh and oh) then
				sep:SetAlpha(0)
			else
				sep:SetAlpha(EventHorizon_Vitals.color.weaponseparator[4])
			end
			
			hTSLU = hTSLU - hTTL
		end
	end
end

frame:SetScript("OnUpdate", EventHorizon_Vitals.OnUpdate)
frame:SetScript("OnEvent", EventHandler)

local talentcheck = GetTalentInfo(1,1)
if talentcheck then
	-- Talent information is already available, which means that the interface was reloaded.
	frame:RegisterEvent('PLAYER_LOGIN')
	function frame:PLAYER_LOGIN()
		EventHorizon_Vitals:CheckEH()
		EventHorizon_Vitals:Init()
		EventHorizon_Vitals:SpawnStyleFrame()
		self:UnregisterEvent('PLAYER_LOGIN')
		gtg = true
			if class == "DEATHKNIGHT" then
				for i=1,6 do
				EventHorizon_Vitals:UpdateRuneType(i)
				end
			end
	end
else
	frame:RegisterEvent('PLAYER_ALIVE')
	function frame:PLAYER_ALIVE()
		EventHorizon_Vitals:CheckEH()
		EventHorizon_Vitals:Init()
		EventHorizon_Vitals:SpawnStyleFrame()
		self:UnregisterEvent('PLAYER_ALIVE')
		gtg = true
			if class == "DEATHKNIGHT" then
				for i=1,6 do
				EventHorizon_Vitals:UpdateRuneType(i)
				end
			end
	end
end

frame:RegisterEvent('UNIT_ENTERED_VEHICLE')
frame:RegisterEvent('UNIT_EXITED_VEHICLE')
function frame:UNIT_ENTERED_VEHICLE(unit)
	if not ((class == "ROGUE")or(class == "DRUID")) then
		frame:RegisterEvent('UNIT_COMBO_POINTS')
		frame:RegisterEvent('PLAYER_TARGET_CHANGED')
	end
	EventHorizon_Vitals:EnterVehicle(unit)
	EventHorizon_Vitals:Init()
end
function frame:UNIT_EXITED_VEHICLE(unit)
	if not ((class == "ROGUE")or(class == "DRUID")) then
		frame:UnregisterEvent('UNIT_COMBO_POINTS')
		frame:UnregisterEvent('PLAYER_TARGET_CHANGED')
	end
	EventHorizon_Vitals:ExitVehicle(unit)
	EventHorizon_Vitals:Init()
end

if ((class == "ROGUE")or(class == "DRUID")) then
	frame:RegisterEvent('UNIT_COMBO_POINTS')
	frame:RegisterEvent('PLAYER_TARGET_CHANGED')
	frame:RegisterEvent('UPDATE_SHAPESHIFT_FORM')
	
	function frame:UNIT_COMBO_POINTS()
		EventHorizon_Vitals:UpdateCP()
	end
	function frame:PLAYER_TARGET_CHANGED()
		EventHorizon_Vitals:UpdateCP()
	end
	function frame:UPDATE_SHAPESHIFT_FORM()
		EventHorizon_Vitals:UpdateCP()
	end
end

if class == "DEATHKNIGHT" then
	frame:RegisterEvent('RUNE_POWER_UPDATE')
	frame:RegisterEvent('RUNE_TYPE_UPDATE')
	
	function frame:RUNE_POWER_UPDATE(runeid,usable)
		EventHorizon_Vitals:UpdateRune(runeid,usable)
	end
	function frame:RUNE_TYPE_UPDATE(runeid)
		EventHorizon_Vitals:UpdateRuneType(runeid)
	end
end
